// Typedefs for my personal preference of seeing how many bits are in the type
typedef unsigned long u32;
typedef signed long s32;
typedef unsigned short u16;
typedef signed short s16;
typedef unsigned char u8;
typedef signed char s8;
typedef u8 UNK;

// Text formatting character codes
const char ALIGN_LEFT = 0x00;
const char ALIGN_CENTER = 0x02;
const char ALIGN_RIGHT = 0x04;

// Values and structs compiled from (some stuff is out of date compared to BWAPI now):
// EUDDB: http://farty1billion.dyndns.org/EUDDB/
// BWAPI: http://code.google.com/p/bwapi/
#pragma pack(1)
namespace RACE {
	enum Enum : u8 {
		Zerg    = 0,
		Terran  = 1,
		Protoss = 2,
		Other   = 3,
		Select  = 5,
		Random  = 6,
		None    = 7,
	};
}

namespace PLAYERTYPE {
	enum Enum : u8 {
		NotUsed          = 0,
		Computer         = 1,
		Human            = 2,
		Rescuable        = 3,
		Unknown0         = 4,
		ComputerSlot     = 5,
		OpenSlot         = 6,
		Neutral          = 7,
		ClosedSlot       = 8,
		Unknown1         = 9,
		HumanDefeated    = 10, /**< Left */
		ComputerDefeated = 11  /**< Left */
	};
}

namespace STATUSFLAGS {
	enum Enum : u32 {
		Completed           = 1,
		OnGround            = 2,
		InAir               = 4,
		Disabled            = 8,
		Burrowed            = 16,
		EnteringBuilding    = 32,
		EnteringTransport   = 64,
		UNKNOWN1            = 128,
		RequiresDetection   = 256,
		Cloaked             = 512,
		DoodadStatesThing   = 1024,
		CloakingForFree     = 2048,
		CanNotReceiveOrders = 4096,
		NoBrkCodeStartThing = 8192,
		UNKNOWN2            = 16384,
		CanNotAttack        = 32768,
		UNKNOWN3            = 65536,
		IsABuilding         = 131072,

		Invincible          = 67108864, // 1 << 17

		SpeedUpgrade        = 268435456, // 1 << 28
		CooldownUpgrade     = 536870912 // 1 << 29
	};
}

struct UNITDEATHS {
	u32				player[12];
};

struct MAPDIMENSIONS {
	u16				width;
	u16				height;
};

struct PLAYER {
	u32				id;
	u32				actions; // unused; FF FF FF FF if not a human player
	u8				type;
	RACE::Enum		race;
	u8				force;
	char			name[25];
};

struct SAVEGROUPS {
	u32				units[18][12];
};

/* vvvvvvvvvvvv Trigger Related vvvvvvvvvvvv */
// See http://www.staredit.net/?p=oldwiki&s=17 for more info on triggers
struct LOCATION {
	u32				topLeftX;
	u32 			topLeftY;
	u32 			bottomRightX;
	u32 			bottomRightY;
	u16 			stringId;
	u16 			flags;
};
struct CONDITION {
	u32				location;
	u32				player;
	u32				number;
	u16				unit;
	u8				comparison;
	u8				condition;
	u8				type;
	u8				flags;
};
typedef BOOL (__fastcall *ConditionPointer)(CONDITION*);
struct ACTION {
	u32				location;
	u32				string;
	u32				wavString;
	u32				time;
	u32				player;
	u32				number;
	u16				unit;
	u8				action;
	u8				number2;
	u8				flags;
};
typedef BOOL (__fastcall *ActionPointer)(ACTION*);
/* ^^^^^^^^^^^^^ Trigger Related ^^^^^^^^^^^^ */

struct ORDER;

typedef struct SPRITE SPRITE;

struct IMAGE { // Length 64
	IMAGE*			previousImage;
	IMAGE*			nextImage;
	u16				imageId;
	u8				paletteType;
	u8				direction;
	UNK unknown1;
	u8				horizontalOffset;
	u8				verticalOffset;
	u16				iscriptHeaderOffset;
	u16				iscriptOffset;
	UNK unknown2[2];
	u8				iscriptCurrentAnimation;
	u8				waitTime;						//?
	u16				currentFrameSet;
	UNK unknown3[19];
	u32				grpOffset;
	u32				overlay;						//?
	u32				renderFunction1;
	u32				renderFunction2;
	SPRITE*			parentSprite;					//?
};

struct SPRITE { // Length 36
	SPRITE*			previousSprite;
	SPRITE*			nextSprite;
	u16				spriteId;
	u8				teamColor;
	UNK unknown0[2];
	u8				animationLevel;
	u8				flags;
	UNK unknown1;
	u16				spriteStructNum;
	UNK unknown2[2];
	u16				currentXPos;
	u16				currentYPos;
	IMAGE*			image;
	IMAGE*			imageUnknown;					//?
	IMAGE*			imageUnknownShadow;				//?
};

struct UNIT { // Length 336 (currently 354...)
	UNIT*			previousUnit;
	UNIT*			nextUnit;
	u32				healthPoints; // HP = healthPoints >> 8 - Internal Counter = healthPoints & 0xFF
	SPRITE*			sprite;
	u16				moveToX;
	u16				moveToY;
	UNIT*			target;
	u16				currentXPos_;
	u16				currentYPos_;
	u16				xPos;
	u16				yPos;
	u8				movementFlags;					//?
	UNK unknown1;
	u8				flingyTurnRadius;
	u8				currentDirection;				//?
	u16				flingyId;
	UNK unknown2;
	u8				flingyMovementType;
	u16				currentXPos;
	u16				currentYPos;
	u32				xHalt;
	u32				yHalt;
	u32				flingySpeed;
	u32				unknownSpeed38;					//?
	u32				unknownSpeed3C;					//?
	UNK unknown3[8];
	u16				flingyAcceleration;				//?
	UNK unknown4[2];
	u8				playerId;
	u8				mainOrderId;
	u8				mainOrderState;
	u8				orderSignal;
	UNK unknown5[4];
	u8				mainOrderTimer;
	u8				groundWeaponCooldown;
	u8				airWeaponCooldown;
	u8				spellCooldown;
	u16				orderTargetX;
	u16				orderTargetY;
	UNIT*			orderTarget;
	u32				shields;
	u16				unitId;
	UNK unknown6[2];
	UNIT*			playerPrevious;
	UNIT*			playerNext;
	UNIT*			subunit;
	ORDER*			orderQueueHead;
	ORDER*			orderQueueTail;
	UNK unknown7[4];
	UNIT*			connectedUnit;
	u8				numberOfQueuedOrders;
	u8				unknownOrderTimer;				//?
	UNK unknown8[2];
	u16				displayedUnitId;				//?
	UNK unknown9[4];
	u8				rankIncrease;
	u8				killCount;
	UNK unknown10[3];
	u8				unknownState;					//?
	u8				currentButtonSet;
	UNK unknown11[3];
	u16				queue[5];
	u16				energy;
	u8				queueSlot;
	u8				counterUnknown; // transport/save group something
	u8				secondaryOrderId;
	UNK unknown13;
	u16				buildRepairHPGain;				//?
	UNK unknown14[2];
	u16				remainingBuildTime;
	UNK unknown15[2];
	u16				transportUnits[8]; // if (transportUnits[n]) UNIT* unit = (UNIT*)(unit_array + 0x150 * ((transportUnits[n] & 0x7FF) - 1));
	union childInfo1Union {
		struct unitIsVulture {
			u8	spiderMineCount;
			UNK	unknown[3];
		}		vultureMines;
		UNIT*	childUnit1;/*
			IF Carrier/Reaver - First Hangar Unit
			IF Scarab/Interceptor - Parent
			IF Building - Addon
			IF Worker - Powerup Carried*/
	}				childInfo1;
	union childInfo2Union {
		struct unitIsBuilding {
			u16	addonBuildId;
			u16	upgradeResearchTime;
		}		buildingInfo;
		UNIT*	nextUnitInParentHanger;/*
			IF Scarab/Interceptor - Next Unit in Parent Hangar*/
	}				childInfo2;
	union childInfo3Union {
		struct unitIsNotChild {
			union scarabsOrTechUnion {
				u8	scarabCount;
				u8	techId;
			}	scarabsOrTech;
			union interceptorsOrUpgradeUnion {
				u8	interceptorCount;
				u8	upgradeId;
			}	interceptorsOrUpgrade;
			u8	larvaSpawnTimer;
			u8	isLanding;
		}		notChildInfo;
		UNIT*	previousUnitInParentHanger;/*
			IF Scarab/Interceptor - Previous in Parent's Hangar*/
	}				childInfo3;
	union childInfo4Union {
		u8	inHanger;
		u8	creepTimer;
		u8	repairingMineralTimer;
	}				childInfo4;/*
		IF Interceptor/Scarab - InHanger?
		IF CreepUnit - timer between creep expansions
		IF Repairing - decrease mineral timer*/
	u8				upgradeLevel;
	u8				isCarryingSomething;			//?
	u8				carryingResourceAmount;
	union unitInfoUnion {
		union subInfoUnion {
			struct isWorkerUnion {
				u16	powerupDropX;
				u16	powerupDropY;
			} isWorker;
			struct isResourceUnion {
				u16	resourceContained;
				u8	resourceIscript;
				u8	resourceCount;				//?
			} isResource;
		}		subInfo;
		UNIT*	resourceOrNydusTarget;
		SPRITE*	nukeDot;
	} 				unitInfo;
	UNK unknown16[8];
	STATUSFLAGS::Enum	status;/*
		00000001(0x1) - Is Completed
		00000010(0x2) - Is on ground? or is it is unit...
		00000100(0x4) - Is in air
		00001000(0x8) - Checked for disabled, if it is 00001000, then the unit is disabled(/unpowered?)
		00010000(0x10) - Checked for burrowing purposes, if it is 00010000, then the unit is burrowed
		00100000(0x20) - Unit is entering a building
		01000000(0x40) - unit is entering a transport
		10000000(0x80) - 
		
		00000001(0x100) - Checked for invisible purposes, if it is 00000001, then the unit requires a detector?
		00000010(0x200) - checked for cloak?
		00000100(0x400) - deals with doodad states? if set, is disabled
		00001000(0x800) - Unit cloaking doesn't need energy decrease
		00010000(0x1000) - Unit is in unbreakable code section? Cannot receive orders
		00100000(0x2000) - Set by nobrkcodestart
		01000000(0x4000) - 
		10000000(0x8000) - cannot attack if set
	
		00000001(0x10000)
		00000010(0x20000) - Is a Building?
	
		00000100(0x4000000) - Invincible
	
		00010000(0x10000000) - Speed upgrade
		00100000(0x20000000) - cooldown upgrade*/
	u8				resourceType;
	u8				wireframeRandomizer;
	u8				secondaryOrderState;
	u8				unknownCountdownE3;				//?
	UNK unknown17[8];
	UNIT*			currentBuildUnit;
	UNK unknown18[8];
	union rallyPsiUnion {
		struct isRally {
			u16	rallyX;
			u16	rallyY;
			UNIT*	rallyUnit;
		}	rallyInfo;
		struct isPsi {
			UNIT*	previousPsiProvider;
			UNIT*	nextPsiProvider;
		}	psiInfo;
	}				rallyPsiInfo;
	u32				pathUnknown;					//?
	UNK unknown19[3];
	u8				isBeingHealed;
	u16				contours1XUnknown;				//?
	u16				contours1YUnknown;				//?
	u16				contours2XUnknown;				//?
	u16				contours2YUnknown;				//?
	u16				removeTimer;/*
		Hallucination, DSwarm, DWeb, Broodling*/
	u16				defenceMatrixDamage;
	u8				defenceMatrixTimer;
	u8				stimTimer;
	u8				ensnareTimer;
	u8				lockdownTimer;
	u8				irradiateTimer;
	u8				statisTimer;
	u8				plagueTimer;
	u8				isUnderStorm;/*
		Used to tell if a unit is under psi storm*/
	UNIT*			irradiatedBy;
	u8				irradiatePlauerId;
	u8				parasiteFlags;/*
		Each bit corrisponds to the player who has parasited this unit*/
	u8				cycleCounter;/*
		Runs updates 2 times per sec(about)*/
	u8				isBlind;
	u8				maelstromCounter;
	UNK unknown20;
	u8				unknownUnreferenced;			//?!
	u8				acidSporeCount;
	u8				acidSporeTime[9];
	u16				offsetIndex3by3;
	UNK unknown21[6];
	u16				airStrength;
	u16				groundStrength;
	UNK unknown22[16];
	u32				tableIdUnknown[4];				//?
	u8				repulse1Unknown;				//?
	u8				repulse2Unknown;				//?
	u8				driftPosX;
	u8				driftPosY;
	UNK unknown23;
};

struct WEAPON {
	WEAPON*			previousWeapon;
	WEAPON*			nextWeapon;
	UNK unknown1[4];
	SPRITE*			sprite;
	u16				targetGroundX;
	u16				targetGroundY;
	UNIT*			targetUnit;
	u8				weaponId;
	UNK unknown2[2];
	u8				bounceCount;
	UNIT*			attackingUnit;
};

//struct BOUNDARY {
//	u16				topLeftX;
//	u16				topLeftY;
//	u16				bottomRightX;
//	u16				bottomRightY;
//};
//
//struct UnitsDATOffsets {
//	u8*				groupFlags;
//	BOUNDARY**		unitBoundaries;
//};

//struct ADDON {
//	u8				xOffset;
//	u8				yOffset;
//};
//struct PLACEMENT {
//	u16				width;
//	u16				height;
//};
//
//struct UNITS_DAT {
//	u8				maxAirHits[228];
//	u16				vespeneCost[228];
//	u8				armor[228];
//	UNK unknown1[4];
//	u16				whatSoundStart[228];
//	u8				compAiUnitType[228];			//?
//	u16				customName[228];
//	u16				buildTime[228];
//	u8				startingDirection[228];
//	u8				broodwarFlag[228];
//	u16				subUnit1[228];
//	u8				spaceProvided[228];
//	u16				restrictionFlags[228];
//	u16				subUnit2[228];
//	u16				shieldAmount[228];
//	u8				movementType[228];
//	u32				constructionAnimation[228];
//	u16				yesSoundEnd[228];
//	u16				starEditAvailabilityFlags[228];
//	u8				airWeapon[228];
//	BOUNDARY		unitBoundaries[228];
//	u16				pissedSoundEnd[106];
//	u16				readySoundStart[106];
//	u8				actionType[228];
//	u8				size[228];
//	u8				humanAiInitAction[228];			//?
//	u32				hitPoints[228];
//	ADDON			addonPosition[192];
//	PLACEMENT		starEditPlacement[228];
//	u16				whatSoundEnd[228];
//	u8				seekRange[228];
//	u8				compAiInitAction[228];			//?
//	u16				portraitId[228];
//	u8				sightRange[228];
//	u8				groundAttackAction[228];		//?
//	u16				buildScore[228];
//	u8				armorUpgrade[228];
//	u8				groundWeapon[228];
//	u8				groupFlags[228];
//	u16				mineralCost[228];
//	u8				airAttackAction[228];			//?
//	u16	unknown2[106];
//	u16	unknown3[228];
//	UNK unknown4[228];
//	UNK unknown5[228];
//	u16				destroyScore[228];
//	u8				specialAbilityFlags[228];
//	u8				transportSize[228];
//	u8				flingyId[228];
//	u8				maxGroundHits[228];
//	u8				supplyProduced[228];
//	u8				shieldsEnabled[228];
//	u8				idleAction[228];				//?
//	u16				infestationUnit[96];
//};
#pragma pack()