#include "MPQDraftPlugin.h"
#include <windows.h>
#include <string>
#include <fstream>
#include <stdio.h>
#include "BW.h"
#include "Settings.h"
//#include "stdafx.h"
//$g++ qdp.cpp -shared -o qdp.dll -Wl,qdp.def
//cl /LD qdp.cpp ..\QDPlugin.def /linkUser32.lib

// A define used for the plugin ID, which is used in the MPQDraft plugin interface 
#define THEPLUGINID 0x40043005; // My method is first digits are version in reverse (0x10xxxxxx = v1, 0x01xxxxxx = v16), other digits are random plugin specific

// A name displayed in the list of MPQDraft/FireGraft plugins
const char* 		plugin_name = "STFPlugin v4 - By poiuy_qwert";

// A function pointer definition for a trigger action (so the function takes a pointer to an ACTION, and returns a BOOL)
typedef BOOL (__fastcall *ActionPointer)(ACTION*);

// Patch Specific globals, mostly from:
// BWAPI: http://code.google.com/p/bwapi/
// BWHacks, now: http://www.gamethreat.net
// EUDDB: http://farty1billion.dyndns.org/EUDDB/
// These are all pointers into SC's code, used for patching functions and other code injections
const u32			BWXFN_GameStart = 0x004C9691; /* 0x004C96F1; /* 1.15.3 */
const u32			BWXFN_GameStartBack = BWXFN_GameStart + 5;
const u32			BWXFN_GameStartTarget = 0x00417C10; /* 0x00417C40; /* 1.15.3 */
const u32			BWXFN_GameEnd = 0x004EE5C3; /* 0x004EE623;/* 1.15.3 */
const u32			BWXFN_GameEndBack = BWXFN_GameEnd + 5;
const u32			BWXFN_GameEndTarget = 0x00416D80; /* 0x00416DB0; /* 1.15.3 */
const u32 			BWXFN_NextLogicFrame = 0x004D936E; /* 0x004D93DE; /* 1.15.3 */
const u32 			BWXFN_NextLogicFrameBack = BWXFN_NextLogicFrame + 5;
const u32 			BWXFN_NextLogicFrameTarget = 0x00488420; /* 0x00488450; /* 1.15.3 */
const u32			BWXFN_Print = 0x004202A0;
const u32			BWXFN_PrintAnywhere = 0x0048CCA0;
const u32			BWXFN_PrintAnywhereBack = BWXFN_PrintAnywhere + 5;
const u32			BWXFN_Refresh = 0x0041E25A; // One null
const u32			BWXFN_RefreshBack = BWXFN_Refresh + 5;
const u32			BWXFN_RefreshRect = 0x0041E0C0;
const u32			BWXFN_CreateUnit = 0x004889F1;
const u32			BWXFN_CreateUnitBack = BWXFN_CreateUnit + 5;
const u32			BWXFN_SelectGroup = 0x004967E6;
const u32			BWXFN_SelectGroupBack = BWXFN_SelectGroup + 5;
const u32			PlayWAV_JNZ = 0x004C7441; /* 0x004C7441; /*1.15.3*/

// These are all pointers into SC's data, used for reading/changing game settings/states
const u32			unit_array = 0x0059CC90;
ActionPointer 		*action_table = (ActionPointer*)0x00512800;
u32					*mouse_position = (u32*)0x0051A3E4;
u32					*elapsed_time = (u32*)0x0058D6E0;
u16					**map_string_offsets = (u16**)0x005993BC;
PLAYER				*player_info = (PLAYER*)0x0057EEE0;
u32					*player_minerals = (u32*)0x0057F0D8;
u32					*screen_x = (u32*)0x00628430;
u32					*screen_y = (u32*)0x00628458;
u16					*unitsdat_gascost = (u16*)0x0065FCE8;
u32					*unitsdat_maxhp = (u32*)0x00662339;
u16					*current_player = (u16*)0x00512678;
SAVEGROUPS			*save_groups = (SAVEGROUPS*)0x0057FE48;
UNIT				**selected_units = (UNIT**)0x006284D0;\

// A Class written for easy file based integer and string settings. See Settings.h/cpp for more details
Settings* settings = NULL;

// A function to play a WAV using the PlayWAV trigger by injecting a wav path into the map strings table
void playWAV(const char* path) {
	const STRING_INDEX = 1;
	// Inject the offset of where the file path is going to be placed
	(*map_string_offsets)[STRING_INDEX] = 0x803;
	// A variable to keep the old protection settings to reset them later
	DWORD OldProt = 0;
	// Change the protection settings at the location the file path will be placed at
	VirtualProtect((char*)(*map_string_offsets) + 0x803, strlen(path) + 1, PAGE_EXECUTE_READWRITE, &OldProt);
	// Copy the file path into the string location
	strcpy((char*)(*map_string_offsets) + 0x803,path);
	// Revert the protection settings
	VirtualProtect((char*)(*map_string_offsets) + 0x803, strlen(path) + 1, OldProt, &OldProt);
	// Create a new trigger ACTION object (Defined in BW.h)
	ACTION *action = new ACTION();
	// Set the string index so it uses the offset to the file path that was injected
	action->wavString = STRING_INDEX;
	// Call the 9th action in the action_table, which is the PlayWAV trigger, using the action just created
	action_table[8](action);
	// Clean up the memory
	delete action;
}

// A global used to track when the user is in a game or not
u8		game_on = 0,
// An array to keep track of a unit specific value
		unit_data[1700]; // Terran = Kill count, Protoss = Energy

// Some things should only be updated every second, so timer_wait is used to compare against elapsed_time to see if a game second has elapsed
u32		timer_wait = 0,
// A variable to keep track of which idle worker to focus on next when tabbing through them
		last_idle = 0,
// A count of the number of currently idle workers
		idle_workers;


// Two enums which describe the indexes of mercs_influence (defined bellow):
//  mercs_influence[InfluenceMerc][InfluenceData]
typedef enum {
	InfluenceMercGunner,
	InfluenceMercLancer,
	InfluenceMercAsher,
	InfluenceMercGrenadier,
	InfluenceMercRanger,
	InfluenceMercSniper,
	
	InfluenceMercCount
} InfluenceMerc;
typedef enum {
	InfluenceDataBaught, // Keeps track of how many of the merc were baught since the last influence update
	InfluenceDataLastIncrease, // Keeps track of the last increase in influence for the merc from the last influence update
	InfluenceDataPlayersBaughtFlags, // Keeps bitflags of which players have baught the merc since the last influce update
	InfluenceDataPlayersBaughtCount, // Keeps track of how many players have baught the merc since the last influce update
	InfluenceDataColor, // Keeps track of the color used to draw the current cost of the merc influence on screen
	InfluenceDataMercCost, // Keeps track of the merc cost
	InfluenceDataAvailable, // Keeps track of if the merc is currently availabe to the current player
	
	InfluenceDataCount
} InfluenceData;

// Two enums which describe the indexes of human_count (defined bellow):
//  human_count[HumanUnit][HumanData]
typedef enum {
	HumanUnitEngineer,
	HumanUnitOilPump,
	HumanUnitSweatShop,

	HumanUnitCount
} HumanUnit;
typedef enum {
	HumanDataUnitCount, // Keeps track of the amount of the unit currently owned
	HumanDataExtra, // Keeps track of an extra unit specific value (displayed in brackets when printed on screen):
						// Engineer keeps track of how many engineers are not inside a building
						// Oil Pump keeps track of how many engineers are inside any oil pump
						// Sweatshop keeps track of how many engineers are inside any sweatshop

	HumanDataCount
} HumanData;

// Variables used to keep track of STF's Human specific stats which are calculated in the nextFrame() function but must be printed from
//  another function at another time.
// The first dimension of mercs_influence is for each Human merc, the second dimension is the data stored for those units
// mercs_influence[merc][0] = baught
//                [merc][1] = last increase
//                [merc][2] = player baught flags
//                [merc][3] = player baught count
//                [merc][4] = color
//                [merc][5] = merc cost
//                [merc][6] = available
u16		mercs_influence[InfluenceMercCount][InfluenceDataCount],
// The first dimension of human_count is for each specific Human unit, the second dimension is the data stored for those units
// human_count[unit][0] = engy/oil pump/sweatshop count
//            [unit][1] = extra
		human_count[HumanUnitCount][HumanDataCount],
// The current projected minerals increase for each player
		minerals[8],
// The current projected influence increase for each player
		gas[8],
// A count of all hubs for the current player
		human_hubs;


// The nextFrame function is called every frame when the user is in a game
void nextFrame() {
	// Reset all variables to 0
	human_hubs = 0;
	idle_workers = 0;
	for (u8 i = 0; i < 8; ++i) {
		if (player_info[i].race == RACE::Terran) {
			minerals[i] = settings->getInt("Human_DefaultMineralIncrease");
		} else {
			minerals[i] = 0;
		}
		gas[i] = 0;
	}
	for (u8 q = 0; q < 3; ++q) {
		human_count[q][HumanDataUnitCount] = 0;
		human_count[q][HumanDataExtra] = 0;
	}
	for (u8 m = InfluenceMercGunner; m <= InfluenceMercSniper; ++m)
		mercs_influence[m][InfluenceDataAvailable] = 0;
	// Create a new action which will be reused for a couple possible actions
	ACTION *action = new ACTION();
	// Loop through all the units in the game to calculate projected incomes and do other unit specific behaviors
	UNIT *unit = NULL;
	for (u16 id = 0; id < 1700; id++) {
		// Calculate the unit pointer. The UNIT struct I used was not fully correct and thus not the correct size, so normal pointer math could not be
		//  used. If someone were to update the unit struct from somewhere like BWAPI they could simply change a unit_array to a UNIT pointer and
		//  just do "unit = unit_array[id];"
		unit = (UNIT*)(unit_array + 0x150 * id);
		// Check that the unit is a valid completed unit for a valid player
		if (unit->unitId <= 228 && player_info[unit->playerId].type && player_info[unit->playerId].type < PLAYERTYPE::Rescuable && unit->status & STATUSFLAGS::Completed) {
			// If the unit is dead reset its unit_data
			if (unit->mainOrderId == 0 || (unit->healthPoints >> 8) <= 0) {
				unit_data[id] = -1;
			// If the unit is not dead and is a Human unit
			} else if (player_info[unit->playerId].race == RACE::Terran) {
				// If the unit is a hub for the current player
				if (unit->unitId == settings->getInt("Human_HubUnitID") && unit->playerId == *current_player) {
					// Increase hub count
					human_hubs++;
				// If the unit is an assault facility for the current player, the gunner and lancer are available
				} else if (unit->unitId == settings->getInt("Human_AssaultFacilityUnitID") && unit->playerId == *current_player) {
					mercs_influence[InfluenceMercGunner][InfluenceDataAvailable] = 1;
					mercs_influence[InfluenceMercLancer][InfluenceDataAvailable] = 1;
				// If the unit is a pyrotechnic facility for the current player, the asher and grenadier are available
				} else if (unit->unitId == settings->getInt("Human_PyrotechnicFacilityUnitID") && unit->playerId == *current_player) {
					mercs_influence[InfluenceMercAsher][InfluenceDataAvailable] = 1;
					mercs_influence[InfluenceMercGrenadier][InfluenceDataAvailable] = 1;
				// If the unit is an espionage facility for the current player, the ranger and sniper are available
				} else if (unit->unitId == settings->getInt("Human_EspionageFacilityUnitID") && unit->playerId == *current_player) {
					mercs_influence[InfluenceMercRanger][InfluenceDataAvailable] = 1;
					mercs_influence[InfluenceMercSniper][InfluenceDataAvailable] = 1;
				// If the unit is an engineer, calculate its mineral gain
				} else if (unit->unitId == settings->getInt("Human_EngineerUnitID")) {
					// If the engineer is the current players
					if (unit->playerId == *current_player) {
						// Increase the engineer count
						human_count[HumanUnitEngineer][HumanDataUnitCount]++;
						// Increase the count of engineers outside of buildings, if the engineer is inside a building it will be subtracted later
						human_count[HumanUnitEngineer][HumanDataExtra]++;
						// If the unit is in an idle state, if the engineer is inside a building it will be subtracted later
						if (unit->mainOrderId == 2 || unit->mainOrderId == 3 || unit->mainOrderId == 4 || unit->mainOrderId == 5 || unit->mainOrderId == 18 || unit->mainOrderId == 160)
							// Increase idle worker count
							idle_workers++;
					}
					// Add the projected mineral amount for an engineer in the open, if the engineer is inside a building it will be subtracted later
					minerals[unit->playerId] += settings->getInt("Human_EngineerInOpenMinGain");
				// If the unit is an oil pump or sweatshop
				} else if (unit->unitId == settings->getInt("Human_OilPumpUnitID") || unit->unitId == settings->getInt("Human_SweatshopUnitID")) {
					// If the unit is an oil pump
					if (unit->unitId == settings->getInt("Human_OilPumpUnitID")) {
						// If the oil pump is the current players
						if (unit->playerId == *current_player)
							// Increase the oul pump count
							human_count[HumanUnitOilPump][HumanDataUnitCount]++;
						// Add the projected influence amount for an oil pump
						gas[unit->playerId] += settings->getInt("Human_OilPumpGasGain");
					// If the unit is a sweatshop
					} else {
						// If the oil pump is the current players
						if (unit->playerId == *current_player)
							// Increase the oul pump count
							human_count[HumanUnitSweatShop][HumanDataUnitCount]++;
						// Add the projected mineral amount for a sweatshop
						minerals[unit->playerId] += settings->getInt("Human_SweatshopMinGain");
					}
					// Loop through the loaded units slots 
					for (u8 n = 0; n < 8; n++) {
						// If the loaded unit slot is not empty
						if (unit->transportUnits[n]) {
							// Calculate the pointer of the loaded unit
							UNIT *loadedunit = (UNIT*)(unit_array + 0x150 * ((unit->transportUnits[n] & 0x7FF) - 1));
							// If the loaded unit is an engineer
							if (loadedunit->unitId == settings->getInt("Human_EngineerUnitID")) {
								// If the engineer is the current players
								if (unit->playerId == *current_player) {
									// Decrease the idle worker count
									idle_workers--;
									// Decrease the count of engineers outside of buildings
									human_count[HumanUnitEngineer][HumanDataExtra]--;
								}
								// Remove the projected minerals gained from the engineer being outside
								minerals[unit->playerId] -= settings->getInt("Human_EngineerInOpenMinGain");
								// If the unit is loaded inside an oil pump
								if (unit->unitId == settings->getInt("Human_OilPumpUnitID")) {
									// Increase the projected influence amount for an engineer in an oil pump
									gas[unit->playerId] += settings->getInt("Human_EngineerInOilPumpGasGain");
									// If the oil pump and engineer are the current players
									if (unit->playerId == *current_player)
										// Increase the count of engineers inside oil pumps
										human_count[HumanUnitOilPump][HumanDataExtra]++;
								// If the unit is loaded inside a sweatshop
								} else {
									// Increase the projected minerals amount for an engineer in a sweatshop
									minerals[unit->playerId] += settings->getInt("Human_EngineerInSweatshopMinGain");
									// If the sweatshop and engineer are the current players
									if (unit->playerId == *current_player)
										// Increase the count of engineers inside of sweatshops
										human_count[HumanUnitSweatShop][HumanDataExtra]++;
								}
							}
							// If a game second has passed
							if (*elapsed_time > timer_wait) {
								// If the unit is loaded inside an oil pump
								if (unit->unitId == settings->getInt("Human_OilPumpUnitID")) {
									// A variable used to calculate hp and or mana regen
									u16 hp_mana = 0;
									// If the elapsed time is on a hp regen interval and the loaded unit is not at max health
									if (!(*elapsed_time % settings->getInt("Human_LoadedHealRate")) && (loadedunit->healthPoints >> 8) < (unitsdat_maxhp[loadedunit->unitId] & 0xFFFF)) {
										// Calculate the resultant hp by adding a percentage of the loaded units max hp
										hp_mana = (loadedunit->healthPoints >> 8) + ((unitsdat_maxhp[loadedunit->unitId] & 0xFFFF) * (settings->getInt("Human_LoadedHealPercent") / 100.0));
										// Ensure not to heal past max health
										if (hp_mana > unitsdat_maxhp[loadedunit->unitId] & 0xFFFF)
											hp_mana = unitsdat_maxhp[loadedunit->unitId] & 0xFFFF;
										// Set the health of the loaded unit
										loadedunit->healthPoints = (loadedunit->healthPoints & 0xFF) + (hp_mana << 8);
									}
									// If the elapsed time is on a mana regen interval and the loaded unit is not at max energy
									if (!(*elapsed_time % settings->getInt("Human_LoadedManaRegenRate")) && (loadedunit->energy >> 8) < settings->getInt("Human_MaxMana")) {
										// Calculate the resultant mana by adding a percentage of the loaded units max energy
										hp_mana = (loadedunit->energy >> 8) + (settings->getInt("Human_MaxMana") * (settings->getInt("Human_LoadedManaRegenPercent") / 100.0));
										// Ensure not to regnerate past max regeneratable amount of energy
										if (hp_mana > settings->getInt("Human_MaxMana"))
											hp_mana = settings->getInt("Human_MaxMana");
										// Set the energy of the loaded unit
										loadedunit->energy = (loadedunit->energy & 0xFF) + (hp_mana << 8);
									}
								// If the elapsed time is on a kills to cash interval and the unit has kills
								} else if (!(*elapsed_time % settings->getInt("Human_LoadedKillsToCashRate")) && loadedunit->killCount) {
									// Setup the ACTION to run a SetResources trigger on for minerals (i use the SetResources trigger instead of
									//  setting the resources manually so there is that nice counting up animation)
									action->number2 = 8; // Add
									action->unit = 0; // Minerals
									// Get X minerals per kill
									action->number = settings->getInt("Human_KillMineralReward") * loadedunit->killCount;
									// Call the SetResources action
									action_table[26](action);
									// Change the resource type to gas
									action->unit = 1; // Gas
									// Get X gas per kill
									action->number = settings->getInt("Human_KillGasReward") * loadedunit->killCount;
									// Call the SetResources action
									action_table[26](action);
									// Set the killcount back to 0
									loadedunit->killCount = 0;
								}
							}
						}
					}
				}
				// If the unit has gained enough kills to get stim count
				if (unit->killCount > 0 && unit->killCount % settings->getInt("Human_KillCountStimKills") == 0 && unit->killCount != unit_data[id]) {
					// Increase the units stim timer
					unit->stimTimer += unit->killCount * settings->getInt("Human_KillCountStimTime");
					// If the unit is on screen
					if (unit->currentXPos >= *screen_x && unit->currentXPos - *screen_x <= 640 && unit->currentYPos >= *screen_y && unit->currentYPos - *screen_y <= 384)
						// Play a sound for the stim
						playWAV(settings->getStr("Human_StimSound"));
				}
				// Keep track of the units killcount
				unit_data[id] = unit->killCount;
			}
		}
	}
	// If a game second has passed
	if (*elapsed_time > timer_wait) {
		// Update timer_wait so we'll know when the next second passes
		timer_wait = *elapsed_time;
		// A boolean to store if the elapsed time is on the mineral compounding interval
		bool addminerals = !(*elapsed_time % settings->getInt("Human_MineralCompoundRate"));
		// A boolean to store if the elapsed time is on the influence compounding interval
		bool addgas = !(*elapsed_time % settings->getInt("Human_GasCompoundRate"));
		// If either resource should be compounded
		if (addminerals || addgas) {
			// Setup the ACTION to run a SetResources trigger 
			action->number2 = 8; // Add
			// Loop through each player
			for (u8 player = 0; player < 8; player += 1) {
				// If its a valid player and a Human player
				if (player_info[player].type && player_info[player].type < PLAYERTYPE::Rescuable && player_info[player].race == RACE::Terran) {
					// Set the player the action is applied to
					action->player = player;
					// If its a mineral compounding interval, the player has minerals to add, and is not at their max stockpile
					if (addminerals && minerals[player] && player_minerals[player] < settings->getInt("Human_MineralStockpileMax")) {
						action->unit = 0; // Minerals
						// Calculate the minerals to add based on current stockpile
						action->number = minerals[player] * ((settings->getInt("Human_MineralStockpileMax") - (int)player_minerals[player]) / (double)settings->getInt("Human_MineralStockpileMax"));
						// Call the SetResource action
						action_table[26](action);
					}
					// If its an influence compounding interval and the player has influence to add
					if (addgas && gas[player]) {
						action->unit = 1; // Gas
						action->number = gas[player];
						// Call the SetResource action
						action_table[26](action);
					}
				}
			}
		}
		// If the elapsed time is on the merc influence update interval
		if (!(*elapsed_time % settings->getInt("Human_MercInfluenceUpdateRate"))) {
			u16 id,inf,increase;
			// Loop through each merc to calculate its new influence cost resulting from their demand
			for (int m = InfluenceMercGunner; m < InfluenceMercCount; m++) {
				// Im not going into to much detail for this. It calculates the next price of the merc
				if (mercs_influence[m][InfluenceDataPlayersBaughtCount] > settings->getInt("Human_MercInfluencePlayerBoughtMaximum")) {
					mercs_influence[m][InfluenceDataPlayersBaughtCount] = settings->getInt("Human_MercInfluencePlayerBoughtMaximum");
				} else if (mercs_influence[m][InfluenceDataPlayersBaughtCount] == 1) {
					mercs_influence[m][InfluenceDataPlayersBaughtCount] = 0;
				}
				increase = mercs_influence[m][InfluenceDataBaught] * (settings->getInt("Human_MercInfluenceIncrease") - (mercs_influence[m][InfluenceDataPlayersBaughtCount] * settings->getInt("Human_MercInfluencePlayerBoughtReduction")));
				if (m == InfluenceMercGunner) {
					id = settings->getInt("Human_GunnerUnitID");
					inf = settings->getInt("Human_GunnerBaseInfluenceCost");
				} else if (m == InfluenceMercLancer) {
					id = settings->getInt("Human_LancerUnitID");
					inf = settings->getInt("Human_LancerBaseInfluenceCost");
				} else if (m == InfluenceMercAsher) {
					id = settings->getInt("Human_AsherUnitID");
					inf = settings->getInt("Human_AsherBaseInfluenceCost");
				} else if (m == InfluenceMercGrenadier) {
					id = settings->getInt("Human_GrenadierUnitID");
					inf = settings->getInt("Human_GrenadierBaseInfluenceCost");
				} else if (m == InfluenceMercRanger) {
					id = settings->getInt("Human_RangerUnitID");
					inf = settings->getInt("Human_RangerBaseInfluenceCost");
				} else if (m == InfluenceMercSniper) {
					id = settings->getInt("Human_SniperUnitID");
					inf = settings->getInt("Human_SniperBaseInfluenceCost");
				}
				increase += mercs_influence[m][InfluenceDataLastIncrease] * (1 - (settings->getInt("Human_MercInfluenceReductionPercent") / 100.0));
				// Set the cost of the merc
				unitsdat_gascost[id] = inf + increase;
				// If the cost went up, the text color will be red
				if (unitsdat_gascost[id] > mercs_influence[m][InfluenceDataMercCost]) {
					mercs_influence[m][InfluenceDataColor] = 6;
				// If the cost stayed the same, the color will be yellow
				} else if (unitsdat_gascost[id] == mercs_influence[m][InfluenceDataMercCost]) {
					mercs_influence[m][InfluenceDataColor] = 3;
				// If the cost went down, the color will be green
				} else {
					mercs_influence[m][InfluenceDataColor] = 7;
				}
				// Reset and keep track of current values
				mercs_influence[m][InfluenceDataBaught] = 0;
				mercs_influence[m][InfluenceDataLastIncrease] = increase;
				mercs_influence[m][InfluenceDataPlayersBaughtFlags] = 0;
				mercs_influence[m][InfluenceDataPlayersBaughtCount] = 0;
				mercs_influence[m][InfluenceDataMercCost] = unitsdat_gascost[id];
			}
		}
	}
	// Delete the ACTION that was created
	delete action;
}

// JmpPatch - From BWAPI by Kovarex
//  This function is used to patch a part of starcrafts code so that it jumps to a function in the plugin. The function that is jumped to must then
//  execute the asm that was written over with the patch, do its stuff, and then jump back right after the patched code at the end of the function
void JmpPatch(void *pDest, int pSrc, int nNops = 0)
{
	DWORD OldProt = 0;
	VirtualProtect((LPVOID)pSrc, 5 + nNops, PAGE_EXECUTE_READWRITE, &OldProt);
	unsigned char jmp = 0xE9;
	memcpy((LPVOID)pSrc, &jmp, 1);
	DWORD address = (DWORD)pDest - (DWORD)pSrc - 5;
	memcpy((LPVOID)(pSrc + 1), &address, 4); 
	for (int i = 0; i < nNops; ++i)
		*(BYTE*)((DWORD)pSrc + 5 + i) = 0x90;
	VirtualProtect((LPVOID)pSrc, 5 + nNops, OldProt, &OldProt);
}

// A function patched into one of starcrafts functions called at game start, used to setup variables for a game
void __declspec(naked) onGameStart()
{
	{
		// Set the game_on global so nextFrameHook knows we're in a game
		game_on = 1;
		// Reset values for the start of a game
		timer_wait = 0;
		for (u16 i = 0; i < 1700; i++)
			unit_data[i] = -1;
		for (u8 player = 0; player < 8; player += 1) {
			if (player_info[player].type && player_info[player].type < PLAYERTYPE::Rescuable)
				player_minerals[player] = settings->getInt("StartingMinerals");
		}
		for (u8 m = InfluenceMercGunner; m < InfluenceMercCount; m++) {
			for (u8 s = 0; s < 4; s++)
				mercs_influence[m][s] = 0;
			if (m == InfluenceMercGunner) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_GunnerBaseInfluenceCost");
			} else if (m == InfluenceMercLancer) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_LancerBaseInfluenceCost");
			} else if (m == InfluenceMercAsher) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_AsherBaseInfluenceCost");
			} else if (m == InfluenceMercGrenadier) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_GrenadierBaseInfluenceCost");
			} else if (m == InfluenceMercRanger) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_RangerBaseInfluenceCost");
			} else if (m == InfluenceMercSniper) {
				mercs_influence[m][InfluenceDataMercCost] = settings->getInt("Human_SniperBaseInfluenceCost");
			}
			mercs_influence[m][InfluenceDataColor] = 5;
			mercs_influence[m][InfluenceDataAvailable] = 0;
		}
	}
	// Call the patched function and then jmp back to where the patch was
	__asm
	{
		call [BWXFN_GameStartTarget]
		jmp [BWXFN_GameStartBack]
	}
}
// A function patched into one of starcrafts functions called at game end, used to reset game_on variable
void __declspec(naked) onGameEnd()
{
	{
		game_on = 0;
	}
	// Call the patched function and then jmp back to where the patch was
	__asm
	{
		call [BWXFN_GameEndTarget]
		jmp [BWXFN_GameEndBack]
	}
}

// A function patched into one of starcrafts functions called every frame
void __declspec(naked) nextFrameHook()
{
	// Call the patched function first, then push all the registers so we can restore them later
	__asm
	{
		// Redo the code that was patched over to jump to this function
		call [BWXFN_NextLogicFrameTarget]
		pushad
	}
	{
		// If the game is on, run the plugin logic in the nextFrame function
		if (game_on)
			nextFrame();
	}
	// Reset all registers and then jump back to where the patch was
	__asm
	{
		popad
		jmp [BWXFN_NextLogicFrameBack]
	}
}

// A function patched into starcrafts "unit was created" function
void __declspec(naked) onCreateUnit() {
	u32 unitId,isValid;
	u8 player;
	__asm {
		// Redo the code that was patched over to jump to this function
		mov ebp,esp
		sub esp,8
		// Save all the registers to be restored later
		pushad
		// Get all the values from the registers into variables
		mov unitId,eax
		mov isValid,ecx
		push edx
		mov dl,byte ptr ds:[edi+0x4C]
		mov player,dl
		pop edx
	}
	{
		// If the unit is valid and the player is valid and Human
		if (isValid == 1 && player < 8 && player_info[player].race == RACE::Terran) {
			s8 m = -1;
			if (unitId == settings->getInt("Human_GunnerUnitID")) {
				m = InfluenceMercGunner;
			} else if (unitId == settings->getInt("Human_LancerUnitID")) {
				m = InfluenceMercLancer;
			} else if (unitId == settings->getInt("Human_AsherUnitID")) {
				m = InfluenceMercAsher;
			} else if (unitId == settings->getInt("Human_GrenadierUnitID")) {
				m = InfluenceMercGrenadier;
			} else if (unitId == settings->getInt("Human_RangerUnitID")) {
				m = InfluenceMercRanger;
			} else if (unitId == settings->getInt("Human_SniperUnitID")) {
				m = InfluenceMercSniper;
			}
			// If the unit is a merc
			if (m > -1) {
				// If this is the first time the player has baught the merc since last influence update
				if (!(mercs_influence[m][InfluenceDataPlayersBaughtFlags] & (1<<player))) {
					// Enable the player baught flag
					mercs_influence[m][InfluenceDataPlayersBaughtFlags] |= (1<<player);
					// Increase players baught count
					mercs_influence[m][InfluenceDataPlayersBaughtCount]++;
				}
				// Increase total baught count since last influence update
				mercs_influence[m][InfluenceDataBaught]++;
			}
		}
	}
	// Restore all registers and jump back to where the patch was
	__asm {
		popad
		jmp [BWXFN_CreateUnitBack]
	}
}

// printText - Written by poiuy_qwert using knowledge from BWHacks
//  This function is a wrapper around starcrafts internal printText function, to make it easier to use the internal function. It prints text (color
//  codes work inside text) wrapped inside a rectangle at x,y with a size of width,height, with specified alignment. Because of the way starcraft
//  rendering is handled, drawing can only happen at specific times without having rendering issues (flickering or parts missing, not being rendered
//  at all at worst). The best place to call it tha i have found is from somewhere that starcraft is calling the internal printText function itself
void printText(char* text, u32 x, u32 y, u32 width=0, u32 height=0x10, char align=ALIGN_LEFT) {
	// Unsure its not printing off screen
	u32 maxx,maxy;
	if (x + width > 640 || width == 0) {
		maxx = 640 - x;
	} else {
		maxx = x + width;
	}
	if (y + height > 480 || height == 0) {
		maxy = 480 - y;
	} else {
		maxy = y + height;
	}
	__asm {
		// Setup all the registers and memmory locations with the print settings
		mov eax,text
		mov esi,x
		mov edi,y
		push edi
		mov ecx,maxy
		mov bl,align
		mov byte ptr ds:[0x6CE0F8],bl
		mov word ptr ds:[0x6CE0B0],si
		mov word ptr ds:[0x6CE0B2],di
		mov ebx,maxx
		mov word ptr ds:[0x6CE0B4],bx
		mov word ptr ds:[0x6CE0B6],cx
		// Call starcrafts printText function
		call [BWXFN_Print]
	}
}

// Check if a point is inside a rectangle
bool inRect(int x,int y,int minx,int miny,int maxx,int maxy) {
	return (x > minx && x < maxx && y > miny && y < maxy ? true : false);
}

// A function patched into one of the places starcraft prints text. The function onRefreshText() later refreshes the screen
void __declspec(naked) onPrintText() {
	// Because all the drawing happens in a place where starcraft is calling the internal printText function itself, starcrafts print settings have to
	//  be saved so they can be restored and reset at the end of the function
	short backup_x,backup_y,backup_maxx,backup_maxy;
	char backup_align;
	__asm {
		// Redo the code that was patched over to jump to this function
		mov al,byte ptr ds:[0x641580]
		// Save all the registers to be restored later
		pushad

		// Save the settings of starcrafts internal print command
		mov al,byte ptr ds:[0x6CE0F8]
		mov backup_align,al
		mov ax,word ptr ds:[0x6CE0B0]
		mov backup_x,ax
		mov ax,word ptr ds:[0x6CE0B2]
		mov backup_y,ax
		mov ax,word ptr ds:[0x6CE0B4]
		mov backup_maxx,ax
		mov ax,word ptr ds:[0x6CE0B6]
		mov backup_maxy,ax
	}
	// If the current player is Human
	if (player_info[*current_player].race == RACE::Terran) {
		// Calculate projected tax
		double tax = (settings->getInt("Human_MineralStockpileMax") - (int)player_minerals[*current_player]) / (double)(settings->getInt("Human_MineralStockpileMax"));
		// Must be between 0 and 1
		if (tax < 0) {
			tax = 0;
		} else if (tax > 1) {
			tax = 1.0;
		}
		// A buffer for dynamicly generated strings
		char buffer[101];
		// Draw projected income information
		printText("\x04Projected Income:",392,16,105,16,ALIGN_RIGHT);
		sprintf(buffer,"\x07$%i (:%i), %ii (:%i)", (u32)(minerals[*current_player] * tax), settings->getInt("Human_MineralCompoundRate") - (*elapsed_time % settings->getInt("Human_MineralCompoundRate")), gas[*current_player], settings->getInt("Human_GasCompoundRate") - (*elapsed_time % settings->getInt("Human_GasCompoundRate")));
		printText(buffer,502,16,120,16);
		// If the mouse is hovering in the economy text area, display the full readout
		if (inRect(mouse_position[1],mouse_position[0],366,15,635,33);) {
			printText("\x04" "Engineers:",392,32,105,16,ALIGN_RIGHT);
			printText("\x04Sweatshops:",392,48,105,16,ALIGN_RIGHT);
			printText("\x04Oil Pumps:",392,64,105,16,ALIGN_RIGHT);
			printText("\x04Taxes:",392,80,105,16,ALIGN_RIGHT);
			printText("\x04Hubs:",392,96,105,16,ALIGN_RIGHT);
			sprintf(buffer,"\x07%i (%i)", human_count[HumanUnitEngineer][HumanDataUnitCount], human_count[HumanUnitEngineer][HumanDataExtra]);
			printText(buffer,502,32,125,16);
			sprintf(buffer,"\x07%i (%i)", human_count[HumanUnitSweatShop][HumanDataUnitCount], human_count[HumanUnitSweatShop][HumanDataExtra]);
			printText(buffer,502,48,125,16);
			sprintf(buffer,"\x07%i (%i)", human_count[HumanUnitOilPump][HumanDataUnitCount], human_count[HumanUnitOilPump][HumanDataExtra]);
			printText(buffer,502,64,125,16);
			sprintf(buffer,"\x07%.2f%%", 100.0 * (1.0 - tax));
			printText(buffer,502,80,125,16);
			sprintf(buffer,"\x07%i", human_hubs);
			printText(buffer,502,96,125,16);
		}
		// 
		if (*elapsed_time > settings->getInt("Human_MercInfluenceDisplayLength") && (*elapsed_time % settings->getInt("Human_MercInfluenceUpdateRate")) < settings->getInt("Human_MercInfluenceDisplayLength")) {
			// If the mouse is hovering in the influence rates area, display the full readout
			if (inRect(mouse_position[1],mouse_position[0],486,294,627,312)) {
				printText("\x04Gunner:",487,215,105,16,ALIGN_RIGHT);
				printText("\x04Lancer:",487,231,105,16,ALIGN_RIGHT);
				printText("\x16" "Asher:",487,247,105,16,ALIGN_RIGHT);
				printText("\x16Grenadier:",487,263,105,16,ALIGN_RIGHT);
				printText("\x04Ranger:",487,279,105,16,ALIGN_RIGHT);
				printText("\x04Sniper:",487,295,105,16,ALIGN_RIGHT);
				for (int m = InfluenceMercGunner; m < InfluenceMercCount; m++) {
					if (mercs_influence[m][InfluenceDataAvailable] == 0) {
						sprintf(buffer,"\x04--");
					} else {
						sprintf(buffer,"%c%i", mercs_influence[m][InfluenceDataColor], mercs_influence[m][InfluenceDataMercCost]);
					}
					printText(buffer,597,215+16*m,30,16);
				}
			// If the mouse is not hovering in the influence rates area, display the condensed readout
			} else {
				for (int m = InfluenceMercGunner; m < InfluenceMercCount; m++) {
					if (mercs_influence[m][InfluenceDataAvailable] == 0) {
						sprintf(buffer,"\x04--");
					} else {
						sprintf(buffer,"%c%i", mercs_influence[m][InfluenceDataColor], mercs_influence[m][InfluenceDataMercCost]);
					}
					printText(buffer,487+26*m,295,20,16);
				}
			}
		}
		// Display idle worker count, in red if there are idle workers
		u8 color = 6;
		// If there are no idle workers, display it in white
		if (idle_workers == 0)
			color = 4;
		sprintf(buffer, "\x04Idle: %c%i", color, idle_workers);
		printText(buffer,30,285,125,16);
		// If the mouse is hovering in the idle worker count area, display a hint on how to select an idle worker
		if (inRect(mouse_position[1],mouse_position[0],30,285,125,317)) {
			sprintf(buffer, "\x04(Press \x07" "0\x04 to select)");
			printText(buffer,30,301,125,16);
		}
	}
	__asm {
		// Restore all of starcrafts internal print settings
		mov al,backup_align
		mov byte ptr ds:[0x6CE0F8],al
		mov ax,backup_x
		mov word ptr ds:[0x6CE0B0],ax
		mov ax,backup_y
		mov word ptr ds:[0x6CE0B2],ax
		mov ax,backup_maxx
		mov word ptr ds:[0x6CE0B4],ax
		mov ax,backup_maxy
		mov word ptr ds:[0x6CE0B6],ax
		// Restore all registers
		popad
		// Jump back to where the patch was
		jmp [BWXFN_PrintAnywhereBack]
	}
}

// A function patched over one of the places starcraft refreshes a part of the screen. Because of the way starcraft rendering is handled, refreshing
//  can only happen at specific times without having rendering issues (flickering or parts missing, not being rendered at all at worst). The best
//  place to call it that i have found is from somewhere that starcraft is calling the internal refresh function itself
void __declspec(naked) onRefreshText() {
	__asm {
		// Save all registers to be restored later
		pushad
		// Setup the paramaters to refresh the entire screen to show our printed text from onPrintText
		mov eax,0
		push 640
		mov ecx,0
		mov edx,480
		// Call starcrafts internal refresh 
		call [BWXFN_RefreshRect]
		// Restore all registers
		popad
		// Jump back to where the patch was
		jmp [BWXFN_RefreshBack]
	}
}

// Used to take a UNIT* and calculate an index stored in selection groups
int calculateIndex(UNIT* unit) {
	return ((((u32)unit - unit_array)/336)|(unit->counterUnknown<<11)) + 1;
}
// A check to see if a unit is not selected
bool notSelected(UNIT* unit) {
	for (int i = 0; i < 12; i++) {
		if (selected_units[unit->playerId*12+i] == unit)
			return false;
	}
	return true;
}
// A check to see if a unit is not loaded
bool notLoaded(UNIT* check_unit) {
	UNIT *unit = NULL;
	for (u16 id = 0; id < 1700; id++) {
		unit = (UNIT*)(unit_array + 0x150 * id);
		if (unit->unitId <= 228 && player_info[unit->playerId].type && player_info[unit->playerId].type < PLAYERTYPE::Rescuable && unit->status & STATUSFLAGS::Completed && (unit->unitId == settings->getInt("Human_OilPumpUnitID") || unit->unitId == settings->getInt("Human_SweatshopUnitID"))) {
			for (u8 n = 0; n < 8; n++) {
				if (unit->transportUnits[n] && (u32)(unit_array + 0x150 * ((unit->transportUnits[n] & 0x7FF) - 1)) == (u32)check_unit)
					return false;
			}
		}
	}
	return true;
}

void bwSelectUnit(void* unitTable, int nUnits)
{
	typedef int (__stdcall* BWSelectUnit)(int nUnits, const void* unitTable);
	BWSelectUnit bwSelectUnit_ = (BWSelectUnit)0x4C0750;
	bwSelectUnit_(nUnits, unitTable);
}

// A place in memory where starcraft stores the last tick control group 0 was pressed. Used to check for double taps
const u32 *last_press = (u32*)0x00597264;
// A function patched into starcrafts "select units in control group" function. To facilitate idle worker selection, when the user presses 0, the next
//  idle worker is placed into control group 0 and starcraft selects it. This was done because at first I didn't think control groups where synced,
//  but unfortunately they are, which causes desyncs in multiplayer. By the time it was brought up I had no time to change it. I hope to eventually
//  be able to but maybe someone else can now that it is release public
void __declspec(naked) onSelectGroup() {
	// Store the control group number and player id
	u32 group,playerId;
	__asm {
		// Get the group number in a variable before replacing it
		mov group,eax
		// Redo the code that was patched over to jump to this function
		mov eax,dword ptr ds:[0x512688]
		// Get the player number in a variable
		mov playerId,eax
		// Save all registers to be restored later
		pushad
	}
	{
		// If its control group 0 and the first time its been pressed in the last 500 ticks
		if (group == 0 && GetTickCount()-*last_press > 500) {
			// Clear control group 0
			for (u8 slot = 0; slot < 12; ++slot)
				save_groups[playerId].units[0][slot] = 0;
			// Start on the unit after the last idle unit slot
			u32 on_unit; = (last_idle + 1) % 1700;
			// Try and find an idle unit
			while (on_unit != last_idle) {
				// Get the unit from the unit array
				UNIT* idle_unit = (UNIT*)(unit_array + 0x150 * on_unit);
				// If the unit is an engineer
				if ((idle_unit->unitId == settings->getInt("Human_EngineerUnitID")) && 
						// And the units owner is valid
						(player_info[idle_unit->playerId].type && player_info[idle_unit->playerId].type < PLAYERTYPE::Rescuable && idle_unit->status & STATUSFLAGS::Completed) &&
						// And the unit is idle
						(idle_unit->mainOrderId == 2 || idle_unit->mainOrderId == 3 || idle_unit->mainOrderId == 4 || idle_unit->mainOrderId == 5 || idle_unit->mainOrderId == 18 || idle_unit->mainOrderId == 160) &&
						// And the unit is not selected
						notSelected(idle_unit) &&
						// And the unit is not loaded
						notLoaded(idle_unit)) {
					// Set the control group to contain the unit
					save_groups[playerId].units[playerId][0] = calculateIndex(idle_unit);
					// Save the unit array index for the next call
					last_idle = on_unit;
					// Break out of the loop
					break;
				}
				// Go to the next unit slot
				on_unit = (on_unit + 1) % 1700;
			}
		}
	}
	__asm {
		// Restore all registers
		popad
		// Jump back to where the patch was
		jmp [BWXFN_SelectGroupBack]
	}
}





// Convenience function to show an error message
void mBox(const char* s) {
	MessageBox(0,s,"Error!",0);
}

// GetSCVersionString - Written by the all powerfull DiscipleOfAdun
//  Returns the string representation (i.e "1.15.1") of the version of an SC exe
char* GetSCVersionString(char* ExePath)
{
	void * data;
	void * verloc;
	DWORD dwVerSize, PVL, PVH, dwExeVerSize;
	char * szVerString;
	//Get the version size
	dwExeVerSize = GetFileVersionInfoSize(ExePath, (unsigned long*)&PVL);
	data = new char[dwExeVerSize];
	if (!GetFileVersionInfo(ExePath, 0, dwExeVerSize, data)) {
		delete[] data;
		return NULL;
	}
	if (!VerQueryValue(data, "\\", &verloc, (PUINT)&dwVerSize)) {
		delete[] data;
		return NULL;
	}
	VS_FIXEDFILEINFO* vers = (VS_FIXEDFILEINFO*)verloc;
	PVH = vers->dwProductVersionMS;
	PVL = vers->dwProductVersionLS;
	int a,b,c,d;
	a = PVH >> 16;
	b = PVH & 0xffff;
	c = PVL >> 16;
	d = PVL & 0xffff;
	szVerString = new char[6];
	if (b < 14) //1.13f or below
		sprintf(szVerString, "%d.%d%d%c", a, b, c, (((d+1)/2)+0x60));
	else
		sprintf(szVerString, "%d.%d.%d", a, b, c);
	return szVerString;
}

// Standard MPQDraft plugin interface. The code below will most likely never change except for CanPatchExecutable(), InitializePlugin(), and
//  TerminatePlugin(). Check MPQDraftPlugin.h for a very detailed description of all the functions
class MPQDraftPluginInterface : public IMPQDraftPlugin {
	HINSTANCE hInstance;
	public:
	BOOL WINAPI Identify(LPDWORD pluginID) {
		if (!pluginID) {
			mBox("No space for plugin id.");
			return false;
		}
		*pluginID = THEPLUGINID;
		return true;
	}
	BOOL WINAPI GetPluginName(LPSTR pPluginName,DWORD namebufferlength) {
		if (!pPluginName) {
			mBox("No buffer supplied for plugin name.");
			return false;
		}
		if (namebufferlength < strlen(plugin_name)) {
			mBox("Name buffer is too short.");
			return false;
		}
		strcpy(pPluginName,plugin_name);
		return true;
	}
	BOOL WINAPI CanPatchExecutable(LPCSTR exefilename) {
		char* version = GetSCVersionString((char*)exefilename);
		// STF is 1.15.1 only. Mods can support multiple versions if the addresses are found for all those versions
		if (!strcmp(version, "1.15.1"))
			return true;
		return false;
	}
	BOOL WINAPI Configure(HWND parentwindow) {
		return true;
	}
	BOOL WINAPI ReadyForPatch() {
		return true;
	}
	BOOL WINAPI GetModules(MPQDRAFTPLUGINMODULE* pluginmodules,LPDWORD nummodules) {
		if (!nummodules)
			return false;
		*nummodules = 0;
		return true;
	}
	BOOL WINAPI InitializePlugin(IMPQDraftServer* server) {
		// Get the starcraft exe file path
		char *exePath = new char[256];
		GetModuleFileName(0, exePath, 256);
		// Get the version of the exe
		char *version = GetSCVersionString(exePath);
		delete[] exePath;
		// If the version of the exe is correct
		if (!strcmp(version, "1.15.1")) {
			// Load the STF settings from STFSettings.txt
			settings = new Settings();
			// Inject code into starcraft to jmp into plugin functions
			JmpPatch(onGameStart, BWXFN_GameStart, 0); // From BWAPI by Kovarex
			JmpPatch(onGameEnd, BWXFN_GameEnd, 0); // From BWAPI by Kovarex
			JmpPatch(nextFrameHook, BWXFN_NextLogicFrame, 0); // From BWAPI by Kovarex
			JmpPatch(onCreateUnit, BWXFN_CreateUnit, 0);
			JmpPatch(onPrintText, BWXFN_PrintAnywhere, 0);
			JmpPatch(onRefreshText, BWXFN_Refresh, 1);
			JmpPatch(onSelectGroup, BWXFN_SelectGroup, 0);
			// To be able to use PlayWAV at will for our uses, an if in sc's code (JNZ) must be removed (NOP'ed)
			// A variable to keep the old protection settings to reset them
			DWORD OldProt = 0;
			// Change the protection settings at the location of the JNZ
			VirtualProtect((LPVOID)PlayWAV_JNZ, 6, PAGE_EXECUTE_READWRITE, &OldProt);
			// Replace the JNZ with NOP's
			for (int i = 0; i < 6; ++i)
				*(BYTE*)(PlayWAV_JNZ + i) = 0x90;
			// Revert the protection settings
			VirtualProtect((LPVOID)PlayWAV_JNZ, 6, OldProt, &OldProt);
			// Exit in success
			return true;
		}
		// Show a version warning
		mBox("Your version of StarCraft is not supported by STF Plugin v4, please downgrade to v1.15.1");
		// Exit in failure
		return false;
	}
	BOOL WINAPI TerminatePlugin() {
		// Delete STF settings on exit
		if (settings != NULL)
			delete settings;
		return true;
	}
	void WINAPI SetInstance(HINSTANCE hInst) {
		hInstance = hInst;
	}
};

MPQDraftPluginInterface thePluginInterface;

// Basic DLL setup
BOOL APIENTRY DllMain( HINSTANCE hInstance, DWORD ul_reason_for_call, LPVOID lpReserved) {
	switch (ul_reason_for_call) {
		case DLL_PROCESS_ATTACH:
			thePluginInterface.SetInstance(hInstance);
			break;
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
	}
	return true;
}

// MPQDraft plugin setup (Note: yes FireGraft uses MPQDraft plugins)
BOOL WINAPI GetMPQDraftPlugin(IMPQDraftPlugin **lppMPQDraftPlugin) {
	*lppMPQDraftPlugin = &thePluginInterface;
	return true;
}